class DXOggMusicManager extends DXOgg;

var() string AmbientIntroOggFile;
var() string AmbientOggFile;
var() string CombatIntroOggFile;
var() string CombatOggFile;
var() string ConversationIntroOggFile;
var() string ConversationOggFile;
var() string OutroOggFile;
var() string DeathOggFile;

var EOggMusicMode musicMode;
var float savedSongPos;
var float musicCheckTimer;
var float musicChangeTimer;

event PostPostBeginPlay()
{
    Super.PostPostBeginPlay();

    if(Len(AmbientOggFile) > 0)
        SetCurrentOgg(AmbientIntroOggFile, AmbientOggFile, 0.0, MTran_Instant);
    musicMode = MUS_Ambient;
}

function Tick(float deltaTime)
{
    local bool bCombat;
    local ScriptedPawn npc;
    local Pawn CurPawn;
    local DeusExLevelInfo info;
    local DeusExPlayer player;
    
    Super.Tick(deltaTime);

    player = DeusExPlayer(GetPlayerPawn());

    if (!player.PlayerIsClient())
    {
        if ((musicMode == MUS_Dying) || (musicMode == MUS_Outro))
            return;
    }
    else
    {
        if (musicMode == MUS_Outro)
            return;
    }
      
    musicCheckTimer += deltaTime;
    musicChangeTimer += deltaTime;

    if (player.IsInState('Interpolating'))
    {
        // don't mess with the music on any of the intro maps
        info = player.GetLevelInfo();
        if ((info != None) && (info.MissionNumber < 0))
        {
            musicMode = MUS_Outro;
            return;
        }

        if (musicMode != MUS_Outro)
        {
            if(Len(OutroOggFile) > 0)
            {
                SetCurrentOgg("", OutroOggFile, 0.0, MTRAN_FastFade);
                musicMode = MUS_Outro;
            }
        }
    }
    else if (player.IsInState('Conversation'))
    {
        if (musicMode != MUS_Conversation)
        {
            if(Len(ConversationOggFile) > 0)
            {
                // save our place in the ambient track
                if (musicMode == MUS_Ambient)
                    savedSongPos = GetCurrentSongPosition();
                else
                    savedSongPos = 0.0;

                SetCurrentOgg(ConversationIntroOggFile, ConversationOggFile, 0.0, MTRAN_Fade);
                musicMode = MUS_Conversation;
            }
        }
    }
    else if (player.IsInState('Dying'))
    {
        if (musicMode != MUS_Dying)
        {
            if(Len(DeathOggFile) > 0)
            {
                SetCurrentOgg("", DeathOggFile, 0.0, MTRAN_Fade);
                musicMode = MUS_Dying;
            }
        }
    }
    else
    {
        // only check for combat music every second

       
        if (musicCheckTimer >= 1.0)
        {
            musicCheckTimer = 0.0;
            bCombat = False;

            // check a 100 foot radius around me for combat
            // XXXDEUS_EX AMSD Slow Pawn Iterator
            //foreach RadiusActors(class'ScriptedPawn', npc, 1600)
            for (CurPawn = Level.PawnList; CurPawn != None; CurPawn = CurPawn.NextPawn)
            {
                npc = ScriptedPawn(CurPawn);
                if ((npc != None) && (VSize(npc.Location - player.Location) < (1600 + npc.CollisionRadius)))
                {
                    if ((npc.GetStateName() == 'Attacking') && (npc.Enemy == player))
                    {
                        bCombat = True;
                        break;
                    }
                }
            }

            if (bCombat)
            {
                musicChangeTimer = 0.0;

                if (musicMode != MUS_Combat)
                {
                    if(Len(CombatOggFile) > 0)
                    {
                        // save our place in the ambient track
                        if (musicMode == MUS_Ambient)
                            savedSongPos = GetCurrentSongPosition();
                        else
                            savedSongPos = 0.0;

                        SetCurrentOgg(CombatIntroOggFile, CombatOggFile, 0.0, MTRAN_FastFade);
                        musicMode = MUS_Combat;
                    }
                }
            }
            else if (musicMode != MUS_Ambient)
            {
                // wait until we've been out of combat for 5 seconds before switching music
                if (musicChangeTimer >= 5.0)
                {
                    if(Len(AmbientOggFile) > 0)
                    {
                        // fade slower for combat transitions
						
                        if (musicMode == MUS_Combat)
                            SetCurrentOgg(AmbientIntroOggFile, AmbientOggFile, savedSongPos, MTRAN_SlowFade);
                        else
                            SetCurrentOgg(AmbientIntroOggFile, AmbientOggFile, savedSongPos, MTRAN_Fade);

                        savedSongPos = 0.0;
                        musicMode = MUS_Ambient;
                        musicChangeTimer = 0.0;
                    }
                }
            }
        }
    }
}

defaultproperties
{
    bHidden=true
}
