/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  OggFile.cpp:  Ogg File Stream

  (c) 2005 Jimbo and Off Topic Productions
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


#include "DXOgg.h"


DWORD WINAPI OggStreamThread(OggStream *stream);


//#define BUF_SIZE  81920  //1
//#define BUF_SIZEF 81920.0f
#define BUF_SIZE  40960  //2
#define BUF_SIZEF 40960.0f


OggStream::OggStream(TCHAR *lpOggNameIntro, TCHAR *lpOggNameLoop)
: oggFile(NULL), oggFileIntro(NULL)
{
    bHasIntro = lpOggNameIntro && *lpOggNameIntro;

    streamVol = 1.0f;

    fileName = lpOggNameLoop;

    //---------------------------
    FString filePath;
    vorbis_info *viIntro;

    if(bHasIntro)
    {
        filePath = oggPath;
        filePath += lpOggNameIntro;

        oggFileIntro = _wfopen((TCHAR*)filePath.GetCharArray().GetData(), TEXT("rb"));

        if(ov_open(oggFileIntro, &vorbisDataIntro, NULL, 0) < 0)
            appErrorf(TEXT("DXOgg: File '%s.ogg' is not a valid ogg morbis file"), (TCHAR*)filePath.GetCharArray().GetData());

        viIntro = ov_info(&vorbisDataIntro, -1);

        if(!ov_seekable(&vorbisDataIntro))
            appErrorf(TEXT("DXOgg: Uhm.....  File '%s.ogg' is not seekable for some..  odd..  reason"), (TCHAR*)filePath.GetCharArray().GetData());

        introLength = ov_time_total(&vorbisDataIntro, -1);
    }

    //---------------------------

    filePath = oggPath;
    filePath += lpOggNameLoop;

    oggFile = _wfopen((TCHAR*)filePath.GetCharArray().GetData(), TEXT("rb"));

    if(ov_open(oggFile, &vorbisData, NULL, 0) < 0)
        appErrorf(TEXT("DXOgg: File '%s.ogg' is not a valid ogg morbis file"), (TCHAR*)filePath.GetCharArray().GetData());

    vorbis_info *vi = ov_info(&vorbisData, -1);

    if(bHasIntro)
    {
        if( (viIntro->channels  != vi->channels) ||
            (viIntro->rate      != vi->rate)     )
        {
            appErrorf(TEXT("DXOgg: File '%s.ogg' does not have the same channels or bitrate as '%s.ogg'"), lpOggNameIntro, lpOggNameLoop);
        }
    }

    if(!ov_seekable(&vorbisData))
        appErrorf(TEXT("DXOgg: Uhm.....  File '%s.ogg' is not seekable for some..  odd..  reason"), (TCHAR*)filePath.GetCharArray().GetData());

    //---------------------------

    WAVEFORMATEX wfe;
    DSBUFFERDESC dsbd;

    IDirectSoundBuffer *baseBuffer;

    appMemzero(&wfe, sizeof(wfe));
    wfe.cbSize = sizeof(WAVEFORMATEX);
    wfe.nChannels = vi->channels;
    wfe.nSamplesPerSec = vi->rate;
    wfe.wBitsPerSample = 16;
    wfe.wFormatTag = 1;
    wfe.nBlockAlign = (wfe.nChannels*wfe.wBitsPerSample)/8;
    wfe.nAvgBytesPerSec = (wfe.nSamplesPerSec*wfe.nBlockAlign);

    dwStreamSize = wfe.nAvgBytesPerSec*5;
    dwStreamSize -= (dwStreamSize%BUF_SIZE);

    float fTime = ((float)wfe.nAvgBytesPerSec)/BUF_SIZEF;
    fTime = (1.0f/fTime)*1000.0f;
    dwStreamDelay = (DWORD)fTime;

    appMemzero(&dsbd, sizeof(dsbd));
    dsbd.lpwfxFormat = &wfe;
    dsbd.dwBufferBytes = dwStreamSize;
    dsbd.dwSize = sizeof(dsbd);
    dsbd.dwReserved = NULL;
    dsbd.dwFlags = DSBCAPS_CTRLPOSITIONNOTIFY | DSBCAPS_CTRLVOLUME | DSBCAPS_GETCURRENTPOSITION2;

    if(dSound->CreateSoundBuffer(&dsbd, &baseBuffer, NULL) != DS_OK)
        appErrorf(TEXT("DXOgg: OggStream: Could not create sound stream buffer"));

    if(baseBuffer->QueryInterface(IID_IDirectSoundBuffer8, (LPVOID*)&buffer) != DS_OK)
        appErrorf(TEXT("DXOgg: OggStream: Could not query the main buffer object"));
    baseBuffer->Release();

    hStreamThread = NULL;
    bPlaying = 0;
    curWritePos = 0;

    //---------------------------

    overBuffer = (LPBYTE)appMalloc(BUF_SIZE, CalhoonSux);
    overBufferSize = 0;

    ResetVolume();

    lpCurVorbisData = bHasIntro ? &vorbisDataIntro : &vorbisData;
}

OggStream::~OggStream()
{
    Stop();

    if(overBuffer)
    {
        appFree(overBuffer);
        overBuffer = NULL;
        overBufferSize = 0;
    }

    if(oggFile)
    {
        ov_closefile(oggFile, &vorbisData);
        oggFile = NULL;
    }

    if(oggFileIntro)
    {
        ov_closefile(oggFileIntro, &vorbisDataIntro);
        oggFileIntro = NULL;
    }
}

void OggStream::SetVolume(float vol)
{
    if(vol > 1.0f) vol = 1.0f;
    if(vol < 0.0f) vol = 0.0f;

    streamVol = vol;

    ResetVolume();
}

void OggStream::ResetVolume()
{
    buffer->SetVolume(ConvertToDBVolume(streamVol*GlobalMusicVolume));
}


void OggStream::Play(float position)
{
    check(buffer);

    DWORD dummy;

    double dPos = double(position);

    if(bHasIntro)
    {
        if(dPos < introLength)
            lpCurVorbisData = &vorbisDataIntro;
        else
        {
            lpCurVorbisData = &vorbisData;
            dPos -= introLength;
        }
    }
    else
        lpCurVorbisData = &vorbisData;

    ov_time_seek(lpCurVorbisData, dPos);

    for(int i=0; i<4; i++)
        UploadStreamSegment();

    hStreamThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)OggStreamThread, (LPVOID)this, 0, &dummy);
}

void OggStream::Stop()
{
    if(bPlaying)
    {
        check(buffer);
        check(hStreamThread);

        bPlaying = 0;
        buffer->Stop();

        if(hStreamThread)
        {
            if(WaitForSingleObjectEx(hStreamThread, 100, 0) == WAIT_TIMEOUT)
                TerminateThread(hStreamThread, 0);

            CloseHandle(hStreamThread);
            hStreamThread = NULL;
        }
    }
}

BOOL OggStream::RestoreBuffer()
{
    check(buffer);

    DWORD status;
    buffer->GetStatus(&status);

    if(status & DSBSTATUS_BUFFERLOST)
    {
        if(buffer->Restore() == DSERR_BUFFERLOST)
            return FALSE;
    }

    return TRUE;
}

float OggStream::GetCurrentPosition()
{
    double curTime = ov_time_tell(lpCurVorbisData);

    if(bHasIntro && (lpCurVorbisData == &vorbisData))
        curTime += introLength;

    return float(curTime);
}


DWORD OggStream::UploadStreamSegment()
{
    LPBYTE lpPCMData = (LPBYTE)appMalloc(BUF_SIZE*4, CalhoonSux);
    int curBitstream = 0, err = 0;
    long BytesRead, TotalBytes=0;

    if(overBufferSize)
    {
        TotalBytes = overBufferSize;
        appMemcpy(lpPCMData, overBuffer, overBufferSize);
    }

    while(TotalBytes < BUF_SIZE)
    {
        BytesRead = ov_read(lpCurVorbisData, (char*)lpPCMData+TotalBytes, BUF_SIZE, 0, 2, 1, &curBitstream);
        if(BytesRead < 0)
            appErrorf(TEXT("DXOgg: Error in bitstream..?"));
        else if(BytesRead == 0)
        {
            //if in an intro, switch to the loop
            if(bHasIntro && (lpCurVorbisData == &vorbisDataIntro))
            {
                curBitstream = 0;
                lpCurVorbisData = &vorbisData;
            }

            ov_raw_seek(lpCurVorbisData, 0);
            BytesRead = ov_read(lpCurVorbisData, (char*)lpPCMData+TotalBytes, BUF_SIZE, 0, 2, 1, &curBitstream);
        }

        TotalBytes += BytesRead;
    }

    //---------------------------------------------

    LPBYTE lpWaveData=NULL, lpWrapWaveData=NULL;
    DWORD dwLockSize, dwWrapLockSize;

    if(curWritePos == dwStreamSize)
        curWritePos = 0;

    buffer->Lock(curWritePos, BUF_SIZE, (LPVOID*)&lpWaveData, &dwLockSize, (LPVOID*)&lpWrapWaveData, &dwWrapLockSize, 0);
    if(dwLockSize)
        appMemcpy(lpWaveData, lpPCMData, dwLockSize);

    if(lpWrapWaveData)
    {
        appMemcpy(lpWrapWaveData, lpPCMData+dwLockSize, dwWrapLockSize);
        curWritePos = dwWrapLockSize;
    }
    else
        curWritePos += dwLockSize;

    buffer->Unlock(lpWaveData, dwLockSize, lpWrapWaveData, dwWrapLockSize);

    //---------------------------------------------

    if(TotalBytes > BUF_SIZE)
    {
        overBufferSize = TotalBytes-BUF_SIZE;
        appMemcpy(overBuffer, lpPCMData+BUF_SIZE, overBufferSize);
    }
    else
        overBufferSize = 0;

    appFree(lpPCMData);

    if(BytesRead < 0)
        appErrorf(TEXT("DXOgg: Argh."));

    return (DWORD)TotalBytes;
}

DWORD WINAPI OggStreamThread(OggStream *stream)
{
    IDirectSoundBuffer8 *buffer = stream->buffer;
    IDirectSoundNotify8 *notify;
    DWORD dwBytes;

    if(stream->bPlaying)
        return 0;

    stream->bPlaying = 1;

    if(buffer->QueryInterface(IID_IDirectSoundNotify8, (LPVOID*)&notify) != DS_OK)
    {
        appErrorf(TEXT("DXOgg: Ogg Stream Thread: Could not query the notify object"));
        return 0;
    }


    BOOL bHasPlayed = 0, bValidEvent = 1;
    DWORD nNotifications = stream->dwStreamSize/BUF_SIZE;
    DWORD dwNextPos=BUF_SIZE;
    HANDLE hEvent = CreateEvent(NULL, 0, 0, NULL);
    DSBPOSITIONNOTIFY *pn = (DSBPOSITIONNOTIFY*)appMalloc(sizeof(DSBPOSITIONNOTIFY)*nNotifications+1, CalhoonSux);

    for(DWORD i=0; i<nNotifications; i++)
    {
        pn[i].dwOffset = (i*BUF_SIZE);
        pn[i].hEventNotify = hEvent;
    }

    pn[nNotifications].dwOffset = DSBPN_OFFSETSTOP;
    pn[nNotifications].hEventNotify = hEvent;

    notify->SetNotificationPositions(nNotifications+1, pn);

    while(stream->bPlaying)
    {
        if(bValidEvent)
            dwBytes = stream->UploadStreamSegment();

        if(!bHasPlayed)
        {
            if(stream->RestoreBuffer())
                buffer->Play(0, 0, DSBPLAY_LOOPING);

            bHasPlayed = 1;
        }

        DWORD event = WaitForSingleObject(hEvent, 10000);

        DWORD status;
        buffer->GetStatus(&status);
        if(!(status & DSBSTATUS_PLAYING))
            break;

        DWORD dwCurPos;
        buffer->GetCurrentPosition(&dwCurPos, NULL);

        if((dwCurPos >= dwNextPos) && (dwCurPos < dwNextPos+BUF_SIZE))
        {
            dwNextPos += BUF_SIZE;
            if(dwNextPos == stream->dwStreamSize)
                dwNextPos = 0;
            bValidEvent = 1;
        }
        else
            bValidEvent = 0;
    }

    buffer->Stop();

    appFree(pn);
    CloseHandle(hEvent);

    if(notify)
        notify->Release();

    stream->bPlaying = 0;

    return 0;
}


